theory GHCTest
imports 
  (* RT: GHCCore already import Main, so no need to reimport again *)
  (* Main *)
  GHCCore
begin

(* Some code to try out the evaluation function *)
fun evaluateN :: "expr \<Rightarrow> nat \<Rightarrow> expr executionResult" where
"evaluateN e 0 = Step e" |
"evaluateN e (Suc v) = (case (evaluateN e v) of
   Step res \<Rightarrow> evaluate res Nil |
   other \<Rightarrow> other)"

(* Code to write Haskell functions more easily in HCore *)
definition mkId where "mkId strlst = Idf (strlst)"
definition mkVar where "mkVar strlst = Var (mkId strlst)"
definition mkLit where "mkLit strlst = Lit (mkId strlst)"
definition mkLitAlt where "mkLitAlt strlst = LitAlt (mkId strlst)"
definition mkDatAlt where "mkDatAlt strlst = DataAlt (mkId strlst)"
definition mkLetNR where "mkLetNR strlst e = Let (NonRec (mkId strlst) e)"
definition mkLetR where "mkLetR lst = Let (Rec (map (apfst mkId) lst))"
definition mkCase where "mkCase e str alts = Case e (mkId str) alts"
definition mkLam where "mkLam str = Lam (mkId str)"
definition mkLams where "mkLams strlst = foldr (\<lambda> str e . Lam (mkId str) e) strlst"
definition mkLamCase where (* Useful for pattern-matching functions *)
"mkLamCase str1 alts = mkLam str1 (mkCase (mkVar str1) (str1@''*'') alts)"
definition mkApp where "mkApp f lst = foldl App (mkVar f) lst"
definition mkLitApp where "mkLitApp f lst = foldl App (mkLit f) lst"
definition mkCons where "mkCons h t = mkLitApp '':'' [h,t]"
definition mkHList where "mkHList lst = foldr mkCons lst (mkLit ''[]'')"

fun mkHNum :: "nat \<Rightarrow> expr" where
"mkHNum (Suc x) = mkLitApp ''S'' [mkHNum x]" |
"mkHNum 0 = mkLit ''Z''"

(* prelude functions *)
definition HCore_drop where
"HCore_drop = (''drop'',mkLamCase ''n''
                                [Alt (mkDatAlt ''S'') [mkId ''n3'']
                                     (mkLamCase ''lst''
                                     [Alt DEFAULT [] (mkLit ''[]'')
                                     ,Alt (mkDatAlt '':'') [mkId ''_'',mkId ''rst'']
                                        (App (App (mkVar ''drop'') (mkVar ''n3'')) (mkVar ''rst''))
                                     ])
                                ,Alt (mkDatAlt ''Z'') [] (mkLam ''x'' (mkVar ''x''))
                                ])"

definition HCore_foldr where
"HCore_foldr = (''foldr'',
  mkLams [''f'', ''z''] (mkLamCase ''lst''
    [Alt DEFAULT [] (mkVar ''z'')
    ,Alt (mkDatAlt '':'') [mkId ''x'',mkId ''xs''] (* f x (foldr f z xs) *)
      (mkApp ''f'' [ mkVar ''x''
                   , mkApp ''foldr'' [mkVar ''f'',mkVar ''z'',mkVar ''xs'']])
    ]))"

definition HCore_compose where
"HCore_compose = (''.'',
  mkLams [''f'', ''g'', ''x''] (mkApp ''f'' [mkApp ''g'' [mkVar ''x'']]))"

definition HCore_id where
"HCore_id = (''id'', (mkLam ''x'' (mkVar ''x'')))"

definition HCore_plus where
"HCore_plus = (''+'', (mkLam ''x'' (mkLamCase ''y''
  [ Alt DEFAULT [] (mkVar ''x'')
  , Alt (mkDatAlt ''S'') [mkId ''py'']
     (mkApp ''+'' [mkLitApp ''S'' [mkVar ''x''], mkVar ''py''])])))"

definition HCore_map where
"HCore_map = (''map'', mkLam ''f'' (mkLamCase ''lst''
  [Alt DEFAULT [] (mkLit ''[]'')
  ,Alt (mkDatAlt '':'') [mkId ''x'',mkId ''xs'']
    (mkLitApp '':'' [mkApp ''f'' [mkVar ''x''],mkApp ''map'' [mkVar ''f'',mkVar ''xs'']])]))"

definition HCore_12List where
"HCore_12List = mkHList [mkLit ''1'',mkLit ''2'']"
value "HCore_12List"
definition HCore_num1 where
"HCore_num1 = App (mkLit ''S'') (mkLit ''Z'')"
definition HCore_num2 where
"HCore_num2 = App (mkLit ''S'') HCore_num1"

value "evaluateN HCore_12List 0"
value "evaluateN (mkLetR [HCore_drop] (App (App (mkVar ''drop'') HCore_num1) HCore_12List)) 15"
value "evaluateN (mkLetR [HCore_drop] (App (App (mkVar ''drop'') HCore_num2) HCore_12List)) 24"

definition HCore_badAssignment where
"HCore_badAssignment = (mkLetNR ''x'' (mkLit ''3'')
                       (mkLetNR ''y'' (mkVar ''x'')
                       (mkLetNR ''x'' (mkLit ''4'') (mkVar ''y''))))"

value "evaluateN HCore_badAssignment 2"
value "makeFresh Nil HCore_badAssignment"
value "evaluateN (makeFresh Nil HCore_badAssignment) 2"

definition Challenge1a where (* foldr (.) id (map drop lengths) xs *)
"Challenge1a xs lengths = mkLetR [HCore_foldr,HCore_compose,HCore_id,HCore_drop,HCore_map]
  (mkApp ''foldr'' [mkVar ''.'', mkVar ''id''
                   ,mkApp ''map'' [mkVar ''drop'',mkHList (map mkHNum lengths)]
                   ,mkHList (map mkHNum xs)])"

definition Challenge1b where (* drop (foldr (+) 0 lengths) xs *)
"Challenge1b xs lengths = mkLetR [HCore_foldr,HCore_compose,HCore_plus,HCore_drop,HCore_map]
  (mkApp ''drop'' [mkApp ''foldr'' [mkVar ''+'', mkLit ''Z'', mkHList (map mkHNum lengths)]
                  ,mkHList (map mkHNum xs)])"

(*
  (* These take a while to compute, so I usually leave them commented
     The numbers are chosen to evaluate until the last "Step" before WHNF is concluded.
  *)

value "evaluateN (Challenge1b [] [0]) 29"
value "evaluateN (Challenge1a [] [0]) 46"

value "evaluateN (Challenge1b [0,0,0,0,1] [1]) 38"
value "evaluateN (Challenge1a [0,0,0,0,1] [1]) 55"

value "evaluateN (Challenge1b [0,0,0,0,1] [1,2]) 87"
value "evaluateN (Challenge1a [0,0,0,0,1] [1,2]) 106"

*)




end
