(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2012-2015, 2017)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2013-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Code_Haskell_Formula_Output
imports  
  CeTA.Ceta
  Version
begin

definition xml_poly_number :: "'a :: show \<Rightarrow> xml" 
  where "xml_poly_number x = XML ''number'' [] [XML_text (show x)]" 

fun xml_monom_list :: "('v :: show)monom_list \<Rightarrow> xml" where 
  "xml_monom_list [] = xml_poly_number (1 :: int)"
| "xml_monom_list ms = (let xmls = map (\<lambda> (x,p). let xx = XML ''variable'' [] [XML_text (show x)]
      in if p = 1 then xx else XML ''power'' [] [xx, xml_poly_number x]) ms in if 
      length xmls = 1 then hd xmls else XML ''product'' [] xmls)" 

lift_definition xml_monom :: "'v :: {show,linorder} monom \<Rightarrow> xml" is xml_monom_list .

fun xml_poly :: "('v :: {show,linorder},'a :: {one,show})poly \<Rightarrow> xml" where 
  "xml_poly [] = xml_poly_number (0 :: int)"
| "xml_poly ms = (let xmls = map (\<lambda> (m,c). if c = 1 then xml_monom m else if m = 1 then xml_poly_number c else
      XML ''product'' [] [xml_monom m, xml_poly_number c]) ms in if length xmls = 1 then hd xmls else XML ''sum'' [] xmls)"

definition my_trace :: "string \<Rightarrow> 'a \<Rightarrow> 'a" where "my_trace s x = x" 

context IA_locale
begin
fun xml_IA_exp where
  "xml_IA_exp (Fun LessF [s,t]) = XML ''less'' [] [xml_IA_exp s, xml_IA_exp t]"
| "xml_IA_exp (Fun LeF [s,t]) = XML ''less_equal'' [] [xml_IA_exp s, xml_IA_exp t]"
| "xml_IA_exp (Fun EqF [s,t]) = XML ''equal'' [] [xml_IA_exp s, xml_IA_exp t]"
| "xml_IA_exp e = xml_poly (IA_exp_to_poly e)"

fun xml_formula :: "('v::{show,linorder}) IA_formula \<Rightarrow> xml" where
  "xml_formula (Atom x) = XML ''atom'' [] [xml_IA_exp x]" 
| "xml_formula (NegAtom x) = XML ''not'' [] [XML ''atom'' [] [xml_IA_exp x]]" 
| "xml_formula (Conjunction x) = XML ''conjunction'' [] (map xml_formula x)" 
| "xml_formula (Disjunction x) = XML ''disjunction'' [] (map xml_formula x)" 

definition show_xml_formula :: "('v::{show,linorder}) IA_formula \<Rightarrow> string" where
  "show_xml_formula x = filter (\<lambda> x. x \<noteq> CHR '' '' \<and> x \<noteq> CHR ''\<newline>'') (show (xml_formula x))" 

lemma export_formulas_code[code]: "check_clause hints \<phi> = my_trace (show_xml_formula \<phi>) (
    let es = map IA_exp_to_poly_constraint (translate_conj (\<not>\<^sub>f \<phi>)) in 
    unsat_checker hints es
       <+? (\<lambda> s. shows ''Could not prove unsatisfiability of IA conjunction '' o shows_nl 
     \<circ> shows_list_gen shows_poly_constraint ''False'' [] '' && '' [] es \<circ> shows_nl \<circ> s))"
  unfolding check_clause_def my_trace_def by simp
end

(*
export_code certify_proof Certified Unsupported Error Inl Inr sumbot version
(* remove, after defining an XML format: *)
  Yes No Terminating Upperbound Nonterminating Confluent Nonconfluent Completed Anything nat_of_integer
  in Haskell module_name Ceta_Formula file "../../generated/Haskell/"
*)

end

