theory Code_Unfolds
  imports Ceta
begin
  
(* get code unfolds *)
ML {*
  (* configuration of output *)
  val show_types = true
  val show_sorts = false
  val eliminate_bound = true
  (* implementation *)
  fun print_type (Type (c,ts)) = if ts = [] then c else let val ss = map print_type ts 
    in c ^ "[" ^ hd ss ^ fold (fn s => fn r => "," ^ s ^ r) (tl ss) "]" end
    | print_type (TVar (n,sort)) = if show_sorts then "(" ^ fst n ^ " :: " ^ @{make_string} sort ^ ")" else fst n 
    | print_type _ = "TFree variables not required"
  fun print_of_type (ty : typ) = if show_types then " : " ^ print_type ty else ""
  fun print_const_type s ty = if show_types then "(" ^ s ^ print_of_type ty ^ ")" else s
  fun print_term (Const (c,ty)) = print_const_type c ty
    | print_term (a $ b) = "(" ^ print_term a ^ ") (" ^ print_term b ^ ")" 
    | print_term (Bound i) = "Bound " ^ string_of_int i
    | print_term (Var (n,ty)) = "Var " ^ print_const_type (fst n) ty
    | print_term (Abs (x,ty,t)) = "\\ " ^ print_const_type x ty ^ ". " ^ print_term 
        (if eliminate_bound then subst_bound (Var ((x,0),ty), t) else t)
    | print_term _ = "Free variables not required"
  fun print_thm (name,thm) = let
    val (l,r) = Thm.prop_of thm |> Logic.dest_equals
    in (name ^ ": " ^ print_term l ^ " -> " ^ print_term r)
    end;
  fun print_code_unfold thy = 
    fold (fn x => fn _ => Output.writeln (print_thm x)) (Code_Preproc.the_code_unfolds thy) ();
  Output.writeln (@{make_string} (length (Code_Preproc.the_code_unfolds @{theory})));
  print_code_unfold @{theory};
  
*}

end
